function [simData] = Simulate(p_sim_in,p_comm,flag_noPlot)
% Main simulation loop.
%
% Input:
%   p_sim_in    simulation parameters, obtained from ConfigureSimulation*()
%               (optional, default: ConfigureSimulation())
%   p_comm      communication parameters, overwriting corresponding
%               parameters in design.*
%               (optional)
%   flag_noPlot flag, when true, no plots are being generated
%               (optional, default: true)
%
%
% Output:
%   simData     struct including results of simulation
%
% --
% Supplementary material (simulation examples) to:
%   Distributed Event-based State Estimation (technical report)
%   Event-based State Estimation: An Emulation-based Approach (journal)
%
% Copyright 2015-17 Max Planck Society. All rights reserved.
% 
% Sebastian Trimpe
% Max Planck Institute for Intelligent Systems
% Autonomous Motion Department
% strimpe(at)tuebingen.mpg.de
%
% Revision history
% [31.03.15]        first version
% [13.04.15, Seb]   Double-check simulation result with CDC14 simulation.
%                   - no noise (since noise generation is different),
%                   impulse disturbance, different agents' initial
%                   estimates
%                   - result basically identical (except numerical things)
%                   => consider sanity-check successful
% [20.04.15, ST]    - Changing structure of p_sys.y_dims/u_dims for greater
%                   flexibility.
%                   - Sanity-check BC simulation: same results before and
%                   after change.
%                   - Change structure input estimate: only up own uhat if
%                   also communicated.
%                   - Implemented and tested thermo-fluid process.
% [??, ST]          Changed communication of inputs: only update own input
%                   if also communicated to others.
% [02.05.15, ST]    Tested BC and thermo-fluid sims, minor fixes.
% [02.05.15, ST]    - Allow simulation parameters as input.
%                   - Seed not managed in this function any more.
% [26.07.15, ST]    Allow for the option to force periodic communication
%                   that can be subsampled from underlying periodic
%                   sampling time. For purposes of comparison (only for 
%                   BC simulation.
% [03.08.15, ST]    Some clean-ups


close all;

%% Input handling
% First input argument:
if nargin<1 || isempty(p_sim_in)
    % no input parameters -> use default
    p_sim = ConfigureSimulation();   
else
    p_sim = p_sim_in;
    %flag_noPlot = true;
end;

% Second input argument:
flag_overwrite_comm = false;    % overwriting communication parameters?
if nargin<2 || isempty(p_comm)
    flag_overwrite_comm = false;
else
    flag_overwrite_comm = true; % overwrite with params in p_comm
end;

% Third input argument:
if nargin<3 || isempty(flag_noPlot)
    flag_noPlot = true;
end;


%% Configuration of simulation example

% Simulation parameters
k_final = p_sim.timeHorz;   % time horizon for simulation

% Which simulation example?
% Load system and perform estimator/controller design.
if p_sim.ex==0
    % Example not supported any more.
    error('Simulation example no longer supported.');
    p_sys = ConfigureSystemCDC14();
    design = DesignControllerAndObserverCDC14(p_sys);
    
elseif p_sim.ex==1
    % Thermo-fluid example, event-based communication
    p_sys = ConfigureSystemThermoFluid();    
    design = DesignControllerAndObserverThermoFluid(p_sys);
    
    if p_sim.flag_forcePeriodic
        error('Periodic sampling not supported for thermo-fluid example.');
    end;
    
elseif p_sim.ex==2
    % Balancing Cube model
    % Two communication types are supported: event-based communication, or
    % periodic communication where period can be a multiple of the system's
    % sampling time Ts.
    p_sys = ConfigureSystemBC();
    
    % What communication design?
    if ~p_sim.flag_forcePeriodic
        % event-based design
        design = DesignControllerAndObserverBC(p_sys);
        
    else
        % periodic communication
        % Recompute gains for resampled system?
        if (~isfield(p_sim,'flag_forcePeriodicNoRedesign') ...
                || ~p_sim.flag_forcePeriodicNoRedesign );
            design = DesignControllerAndObserverBC(p_sys,p_sim.forcedPeriod);
        end;
    end;
    
else
    % Unknown
    error('Unknown example.');
    
end;

% Should the communication settings in design.* be overwritten with those
% from input argument p_comm?
if flag_overwrite_comm
    % overwrite communication parameters with those in p_comm (input arg)
    design.EB_y = p_comm.EB_y;
    design.EB_u = p_comm.EB_u;
    design.syncAvg = p_comm.syncAvg;
end;

% Model of system (used in estimator)
p_model = p_sys;    % Assume true model known; model uncertainty modeled 
                    % by process noise.


%% Variables
% Some dimensions
N = p_sys.numAgents;        % number of agents
n_u = p_sys.n_u;            % signal dimensions
n_x = p_sys.n_x;
n_y = p_sys.n_y;
n_x_e = p_sys.est.n_x;      % signal dimensions (for estimation). For 
                            % estimation, a reduced order model may be used,
                            % thus we distinguish n_x and n_x_e.

% Struct carrying system variables (state, measurement, and input)
syst.x = p_sys.x_init;          % state x(k)
syst.y = [];                    % measurement y(k)
syst.u = zeros(n_u,1);          % currently applied input u(k-1)

% Initialize agents
agent = InitAgents(p_sys,design,p_sim.flag_randInitEst);

% Struct carrying centralized estimator states (for comparison)
estC.xhat = zeros(n_x_e,1);
estC.u_minus = zeros(n_u,1);
estC.L = design.L;
estC.xhat(:,1) = p_sys.est.x_init;

% Struct to save simulation data
simData.x = zeros(k_final,n_x);
simData.u = zeros(k_final,n_u);
simData.y = zeros(k_final,n_y);
simData.gamma_syncAvg = nan(k_final,1);     % keep track of when sync averaging happens

% Initialize input vector (for first step)
u = zeros(n_u,1);


%% Main loop
% Simulate over k_final steps.
%
% The main loop repeats the following steps (cf. Alg. 1 in paper):
%   - System update
%   
%   - State estimation: prediction step
%   - Event-based communication of sensor measuremenst
%   - State estimation: measurement update step
%   - Synchronous averaging of agents' estimates (at large period)
%   
%   - Control: compute local control input
%   - Event-based communication of inputs

for k = 1:k_final
    
    %%%%%%%%%%%%%%%%%
    % System update
    %%%%%%%%%%%%%%%%%
   
    % System update step
    syst = SystemUpdate(k,syst,u,p_sys,p_sim);
    
    % Agents' measurements:
    for i=1:N
        for iS=1:agent(i).nSens     % one agent may have more than 1 sensor
            agent(i).y{iS} = syst.y(p_sys.y_dims{i}{iS});
        end;
    end;

    
    
    %%%%%%%%%%%%%%%%%%%%
    % State estimation 
    %%%%%%%%%%%%%%%%%%%%
    
    % Prediction
    for i=1:N
        agent(i) = EstimatorPrediction(agent(i),p_model.est);
    end;
    
    % Event-based communication of measurements
    % Initilize set of received measurements (all empty)
    for i=1:N
        agent(i).y_rcv = [];
    end;
    %    
    for i=1:N   % for each agent
        for iS=1:agent(i).nSens     % for each sensor
            % Compute transmit decision
            agent(i).diff_y{iS} = norm(agent(i).y{iS} - agent(i).C_i{iS}*agent(i).xpred);
            agent(i).gamma_y(iS) = (agent(i).diff_y{iS} >= agent(i).delta_y{iS}) || p_sim.flag_fullCommMeas;
            
            % Force periodic communication?  If yes, overwrite transmit
            % decision.  (This is for the purpose of comparison and not
            % part of the usual event-based control loop.)
            if p_sim.flag_forcePeriodic
                % Periodic communication activated. Overwrite gamma.
                % Enforced periodic communication at period 
                % Ts*p_sim.forcedPeriod.
                agent(i).gamma_y(iS) = 0;
                agent(i).gamma_y(iS) = (mod(k,p_sim.forcedPeriod)==0);  
            end;
            
            % If triggered, communicate measurement to all agents (add to
            % y_rcv)
            if agent(i).gamma_y(iS)
                for j=1:N
                    % Random packet drop:
                    % Packet is lost if
                    %   - packet drops enabled AND
                    %   - packet is dropped (random event) AND
                    %   - not the local agent (no communication required
                    %   for this one)
                    if ( p_sim.flag_packetDrop ...
                            && (rand<p_sim.packetDrop.prob) ...
                            && j ~= i )
                        % Packet dropped
                        % Do nothing.
                    else
                        % Packet received at agent j
                        
                        % Add measurement to set of received measurements.
                        % Format: [y_i, i; y_j, j; ...]
                        agent(j).y_rcv = [agent(j).y_rcv; ...
                            agent(i).y{iS}, agent(i).indices_y{iS}];
                    end;
                end;
            end;
        end;
    end;
    clear iS;
    
    % Measurement update
    for i=1:N
        agent(i) = EstimatorUpdate(agent(i),p_model.est);
    end;
    
    % Synchronous ageraging
    if p_sim.flag_syncAvg && mod(k,design.syncAvg.K)==1
        % Keep track that we average at this step.
        gamma_syncAvg = 1;
        
        % Compute averages
        xhat_avg = zeros(n_x_e,1); 
        xhat_full_avg = zeros(n_x,1);
        for i=1:N
            xhat_avg = xhat_avg+agent(i).xhat;  % estimate of reduced state vector
            xhat_full_avg = xhat_full_avg+agent(i).xhat_full;   % estimate of full state vector
        end;
        xhat_avg = xhat_avg/N;
        xhat_full_avg = xhat_full_avg/N;
        
        % Set agents to averages
        for i=1:N
            agent(i).xhat = xhat_avg;
            agent(i).xhat_full = xhat_full_avg;
        end;
    else
        gamma_syncAvg = 0;
    end;
    
    % Central estimator (for comparison)
    % In normal mode (e.g. when p_sim.flag_centralizedControl is not set)
    % the centralized estimator is used for comparison only.
    estC = CentralEstimator(estC,syst.u,syst.y,p_model.est);
    
    
    
    %%%%%%%%%%%
    % Control
    %%%%%%%%%%%
    
    % Compute local controls (distributed, on each agent)
    for i=1:N
        agent(i) = LocalControl(agent(i));
    end;
    
    % Event-based communication of inputs
    for i=1:N   % for each agent
        for iA=1:agent(i).nInp  % for each actor (group of inputs)
            iu = p_sys.u_dims{i}{iA};   % index of this input
                                        % For simplicity, we assume that 
                                        % all actuators are scalar.
            
            % Communicate to other agents?  Compute transmit decision.
            agent(i).diff_u{iA} = norm(agent(i).u{iA} - agent(i).u_last{iA});
            agent(i).gamma_u(iA) = (agent(i).diff_u{iA} >= agent(i).delta_u{iA}) || p_sim.flag_fullCommInputs;
            
            % Force periodic communication?  If yes, overwrite transmit
            % decision.  (This is for the purpose of comparison and not
            % part of the usual event-based control loop.)
            if p_sim.flag_forcePeriodic
                % periodic communication activated. Overwrite gamma.
                agent(i).gamma_u(iA) = 0;
                agent(i).gamma_u(iA) = (mod(k,p_sim.forcedPeriod)==0);
            end;
            
            if (agent(i).gamma_u(iA))
                % Communicate input to all other agents.
                % Note: The agent updates its own estimate uhat, only if it
                % is also communicated.
                % Note: input communication is assumed perfect, i.e. no
                % packet drops.
                for j=1:N
                    agent(j).uhat(iu) = agent(i).u{iA};
                end;
                
                % Update last communicated input
                agent(i).u_last{iA} = agent(i).u{iA};
            else
                % Don't communicate
                % Agents keep last estimate.
            end;
        end;
    end;
    clear iA;
    

    %%% Control vector for next step
    % Control vector: construct from agents' individual inputs
    u = zeros(p_sys.n_u,1);
    for i=1:N   % for each agent
        for iA=1:agent(i).nInp  % for each actuator
            iu = p_sys.u_dims{i}{iA};
            u(iu) = agent(i).u{iA};
        end;
    end;
    if p_sim.flag_centralizedControl
        % if enabled, overwrite control (for development, testing)
        %u = design.F*syst.x;   % true state
        u = design.F*estC.xhat_full;   % state estimate by centralized
                                        % observer
    end;
        
    
    %%% Store data
    % Update simData struct
    simData = UpdateSimData(simData,k,syst,agent,estC,u,gamma_syncAvg);
    
end;

% Add some parameters etc. to results struct
simData.params.p_sim = p_sim;
simData.params.p_sys = p_sys;
simData.params.design = design;


%% Plot
if ~flag_noPlot
    PlotSimDataP(simData,p_sim.ex);
    % PlotSimData(simData,p_sim.ex);    % for development (ST: keep this)
end;