function design = DesignControllerAndObserverBC(p_sys,subsampleFactor)
% Design the centralized state observer and the
% state-feedback controller.
%
% Input:
%   p_sys       system model parameters
%   subsampleFactor (optional) Resampling of discrete-time model.
%                   Default: [] (empty)
%                   If set, this allows to resample the discrete-time model
%                   at a larger sampling time Ts.  This can be used to
%                   obtain periodic sampling at larger periods than the
%                   base period (used for purposes of comparison).
%
% Output:
%   design      struct including controller and estimator design
%
% --
% Supplementary material (simulation examples) to:
%   Distributed Event-based State Estimation (technical report)
%   Event-based State Estimation: An Emulation-based Approach (journal)
%
% Copyright 2015-17 Max Planck Society. All rights reserved.
% 
% Sebastian Trimpe
% Max Planck Institute for Intelligent Systems
% Autonomous Motion Department
% strimpe(at)tuebingen.mpg.de
%
% Revision history
% [01.04.15, ST]    first version
% [26.07.15, ST]    Added option: allow to subsample system by factor
%                   SUBSAMPLEFACTOR and compute gains accordingly
% [03.08.15, ST]    some clean up


%% Program control
flag_verbose = false;   % true: display some results

% Input handling
if nargin<2 || isempty(subsampleFactor)
    subsampleFactor = [];
end;


%% Central State-feedback controller: LQR design
% Design according to:
% Trimpe/D'Andrea, "The Balancing Cube: A Dynamic Sculpture as Test Bed 
% for Distributed Estimation and Control", Control Systems Magazine, Dec.
% 2012.  
% Trimpe/D'Andrea, "Numerical models and controller design parameters for
% the Balancing Cube"

% Full model
% Control design is done with the full model, including the previous input
% as states.
% States: [modPos(1:6), modVel(1:6), cubePos, cubeVel]
A_full = p_sys.A;
B_full = p_sys.B;

% Resample the system with a longer sample time (Ts*subsampleFactor).
% Used for purpose of comparison of slow periodic with event-based
% sampling.
if ~isempty(subsampleFactor)
    % resample system
    disp('BC: Resample model for control design.');
    
    sys_d1 = ss(A_full, B_full, eye(length(A_full)), [], p_sys.Ts);
    sys_d2 = d2d(sys_d1, p_sys.Ts*subsampleFactor);
    [A_,B_,C_,D_,Ts_] = ssdata(sys_d2);
    
    A_full = A_;
    B_full = B_;
end;

% Weight matrices (from above report, Sec. 4.1), without integrators.
Q_modPos = eye(6);
Q_cube = zeros(2);
R_e = 0.035*eye(6);
Theta_e = 4*eye(6);

Q = blkdiag(Q_modPos, Theta_e, Q_cube);
R = R_e+Theta_e;
N = [zeros(6,6); -Theta_e; zeros(2,6)];

% % Design corresponding to CDC14 paper (for development/debugging):
% % Weight matrices
% R = 0.1*eye(6);
% Q = diag([ones(1,6), zeros(1,6), 1 1]);
% N = [];


% Discrete time LQR design.
[F,S,e] = dlqr(A_full,B_full,Q,R,N);

% Controller gain:
F = -F;     % positive feedback


% Some outputs
if flag_verbose
    disp('LQR design:')
    disp(' ');
    disp('F:');
    disp(F);
    disp('Closed-loop eigenvalues:');
    disp(eig( A_full+B_full*F ));
end;

% Result
design.closedLoopDynCtrl = A_full+B_full*F;
design.F = F;


%% Central Luenberger Observer
% Designed as steady-state Kalman filter.

% Model used for observer design
A_o = p_sys.est.A;
C_o = p_sys.est.C;

% Resample the system with a longer sample time (Ts*subsampleFactor).
% Used for purpose of comparison of slow periodic with event-based
% sampling.
if ~isempty(subsampleFactor)
    % resample system
    disp('BC: Resample model for observer design.');
    
    clear sys_d1 sys_d2;
    sys_d1 = ss(A_o, p_sys.est.B, C_o, [], p_sys.Ts);
    sys_d2 = d2d(sys_d1, p_sys.Ts*subsampleFactor);
    [A_,B_,C_,D_,Ts_] = ssdata(sys_d2);
    
    A_o = A_;
    C_o = C_;
end;


%%% Design III
% Design that is unstable without sync averaging.
%
Qn = 0.1 * diag([ones(1,6), 0*1e-6, 0*1e-6]);
Rn = eye(9);                        % measurement noise

% Steady-state KF design
P = dare(A_o',C_o',Qn,Rn)';
L = P*C_o'*inv(C_o*P*C_o'+Rn);


% Previous/alternative designs: (development, ST: keep)
%
% %%% Design I
% % Noise matrices -> tuning parameters.
% % Baseline: Tuning used in NecSys12-Paper (see.
% % EstimatorCubeEdge_TVKF_Design005.m).
% %   spec.Q = 0.1*diag([1*ones(1,12), 0.2, 1]);  
% %   spec.Rdiag = repmat([0.1, 1],1,6)'; 
% Qn = 0.1 * diag([ones(1,6), 0.2, 1]);
% Rn = diag([ 0.1, 1,  0.1, 1,  0.1, 1,  1, 1, 1]);
%
%%% Design II
% Design that is unstable without sync averaging.
%
% % Design corresponding to CDC14 (still stable)
% Qn = 0.1 * diag([ones(1,6), 0.001, 0.001]);
% Rn = eye(9);                        % measurement noise
%
% Making weights Q(7,7), Q(8,8) small, causes instability (estimator
% dynamics seem to be more coupled then).
% -> Design III


% Some analysis
if flag_verbose
    disp('Observer design:')
    disp(' ');
    disp('L:');
    disp(L);
    disp('Closed-loop eigenvalues:');
    disp(eig( (eye(length(A_o))-L*C_o)*A_o ));
end;

% Save results
design.L = L; 
design.closedLoopDynEst = (eye(length(A_o))-L*C_o)*A_o;


%% Event-based communication of measurements
% Transmit thresholds, one for each sensor.  EB_y.delta should have the 
% same structure as p_sys.y_dims.
delta_enc = 0.001; %0.005;
delta_gyro = 0.004;
design.EB_y.delta{1} = {delta_enc, delta_gyro};
design.EB_y.delta{2} = {delta_enc, delta_gyro};
design.EB_y.delta{3} = {delta_enc, delta_gyro};
design.EB_y.delta{4} = {delta_enc};
design.EB_y.delta{5} = {delta_enc};
design.EB_y.delta{6} = {delta_enc};

                
%% Synchronous averaging of estimates
design.syncAvg.K = 200;


%% Event-based communication of inputs
% Same convention as for measurements.
delta_u = 0.01; % 0.05
design.EB_u.delta{1} = {delta_u};
design.EB_u.delta{2} = {delta_u};
design.EB_u.delta{3} = {delta_u};
design.EB_u.delta{4} = {delta_u};
design.EB_u.delta{5} = {delta_u};
design.EB_u.delta{6} = {delta_u};
